function [cost,grad,costSet] = sparseAutoencoderCost(theta,visibleSize,hiddenSize,...
    lambda,mu,sparsityParam,beta,data,L,gamma)

% visibleSize: the number of input units (probably 64) 
% hiddenSize: the number of hidden units (probably 25) 
% lambda: weight decay parameter
% sparsityParam: The desired average activation for the hidden units (denoted in the lecture
%                           notes by the greek alphabet rho, which looks like a lower-case "p").
% beta: weight of sparsity penalty term
% data: Our 64x10000 matrix containing the training data.  So, data(:,i) is the i-th training example. 
  
% The input theta is a vector (because minFunc expects the parameters to be a vector). 
% We first convert theta to the (W1, W2, b1, b2) matrix/vector format, so that this 
% follows the notation convention of the lecture notes. 

W1 = reshape(theta(1:hiddenSize*visibleSize), hiddenSize, visibleSize);
W2 = reshape(theta(hiddenSize*visibleSize+1:2*hiddenSize*visibleSize), visibleSize, hiddenSize);
b1 = theta(2*hiddenSize*visibleSize+1:2*hiddenSize*visibleSize+hiddenSize);
b2 = theta(2*hiddenSize*visibleSize+hiddenSize+1:end);

% Cost and gradient variables (your code needs to compute these values). 
% Here, we initialize them to zeros. 
cost = 0;
W1grad = zeros(size(W1)); 
W2grad = zeros(size(W2));
b1grad = zeros(size(b1)); 
b2grad = zeros(size(b2));

%% ---------- YOUR CODE HERE --------------------------------------
%  Instructions: Compute the cost/optimization objective J_sparse(W,b) for the Sparse Autoencoder,
%                and the corresponding gradients W1grad, W2grad, b1grad, b2grad.
%
% W1grad, W2grad, b1grad and b2grad should be computed using backpropagation.
% Note that W1grad has the same dimensions as W1, b1grad has the same dimensions
% as b1, etc.  Your code should set W1grad to be the partial derivative of J_sparse(W,b) with
% respect to W1.  I.e., W1grad(i,j) should be the partial derivative of J_sparse(W,b) 
% with respect to the input parameter W1(i,j).  Thus, W1grad should be equal to the term 
% [(1/m) \Delta W^{(1)} + \lambda W^{(1)}] in the last block of pseudo-code in Section 2.2 
% of the lecture notes (and similarly for W2grad, b1grad, b2grad).
% 
% Stated differently, if we were using batch gradient descent to optimize the parameters,
% the gradient descent update to W1 would be W1 := W1 - alpha * W1grad, and similarly for W2, b1, b2. 
% 

[lenPatch,numPatch] = size(data);
X = data;
z2 = W1*X+b1*ones(1,numPatch);
a2 = sigmoid(z2);
z3 = W2*a2+b2*ones(1,numPatch);
a3 = sigmoid(z3);

rho = sparsityParam;
rho_hat = sum(a2,2)/numPatch;

cost1 = 0.5*1/numPatch*sum(sum((a3-X).^2,2));
cost2 = lambda*(sum(sqrt(sum(W1.^2,1))));
cost3 = mu/2*sum(sum(W2.^2));
cost4 = gamma*trace(a2*L*a2');
cost5 = beta*sum(KL(rho,rho_hat));

cost = cost1+cost2+cost3+cost4+cost5;

delta3 = (a3-X).*a3.*(1-a3);
delta2 = (W2'*delta3+beta*(-rho./rho_hat+(1-rho)./(1-rho_hat))*ones(1,numPatch)).*...
    a2.*(1-a2); % Second term in bracket is due to sparsity

DW1 = zeros(1,size(W1,2));
for j = 1:size(W1,2)
    if norm(W1(:,j)) ~= 0
        DW1(j) = 1/norm(W1(:,j));
    end
end

W1grad = delta2*X'/numPatch + lambda*W1*diag(DW1) + gamma*2*a2*L.*a2.*(1-a2)*X';
W2grad = delta3*a2'/numPatch + mu*W2;
b1grad = sum(delta2,2)/numPatch + gamma*sum(2*a2*L.*a2.*(1-a2),2);
b2grad = sum(delta3,2)/numPatch;

%-------------------------------------------------------------------
% After computing the cost and gradient, we will convert the gradients back
% to a vector format (suitable for minFunc).  Specifically, we will unroll
% your gradient matrices into a vector.

grad = [W1grad(:);W2grad(:);b1grad(:);b2grad(:)];

end