clc
close all
clear all

% Parameter Setup (adjustable)
hiddenSize = 10; % single-layer autoencoder hidden layer size
lambda = 1e-2; % parameter that controls column sparsity on W1
gamma = 1e-3; % parameter that controls local data structure preservation
numFeat = 150; % number of features used for classification and clustering
numNN = 5; % number of neighbors for kNN graph construction

% Load both training and testing sets
% MNIST is used for demonstration
% 1000 samples per class are randomly selected for training to alleviate
% negative influence from imbalanced class
disp('load data...')
load('MNIST_train_subset.mat')
load('MNIST_test.mat')

% Calculate sample numbers in both training and testing sets
disp('Calculate sample numbers in both training and testing sets ...')
numTrSubset = length(labTrSubset);
numTe = length(labTe);

% Construct a k-NN graph
disp('construct a k-nn graph')
SkNN = kNNGraph(numNN);
degs = sum(SkNN, 2);
L = diag(degs) - SkNN;
% avoid dividing by zero
degs(degs == 0) = eps;
% calculate D^(-1/2)
DInvSqrt = spdiags(1./(degs.^0.5),0,length(degs),length(degs));
% calculate normalized Laplacian
L = DInvSqrt * L * DInvSqrt;

% Autoencoder training
disp('train an autoencoder...')
visibleSize = size(dataTrSubset,1);
[sae1OptTheta,fAll] = stackedSAEUnsup(dataTrSubset,hiddenSize,400,SkNN,...
    degs,L,lambda,0,0.1,gamma,0);

% Feature Selection
disp('select features ...')
W1 = reshape(sae1OptTheta(1:hiddenSize*visibleSize),hiddenSize,visibleSize);
W1sum1 = sum(abs(W1.^2),1);
disp('show W1 ...')
figure,imagesc(abs(W1)),colorbar % display W1
[val,ind] = sort(W1sum1,'descend');
relativeVal = val/max(val);
featKeepInd = zeros(1,numFeat);
featKeepInd = ind(1:numFeat);
featKeepRatio = numFeat/visibleSize;

% Classification on selected features (softmax)
disp('Classification on selected features (softmax) ...')
numCls = length(unique(labTrSubset));
softmaxLambda = 0;
softoptions = struct;
softoptions.maxIter = 400;
dataTrFeatSelect = dataTrSubset(featKeepInd,:);
dataTeFeatSelect = dataTe(featKeepInd,:);
labTrSubset(find(labTrSubset==0)) = 10; % set digit 0 as the 10th class for training
labTe(find(labTe==0)) = 10; % set digit 0 as the 10th class for evaluation
softmaxModel = softmaxTrain(numFeat,numCls,softmaxLambda,...
    dataTrFeatSelect,labTrSubset,softoptions);
saeSoftmaxOptTheta = softmaxModel.optTheta(:);
softmaxTheta = reshape(saeSoftmaxOptTheta(1:numFeat*numCls),numCls,numFeat);
[~, pred] = max(softmaxTheta * dataTeFeatSelect);
errCls = zeros(1,numCls);
for i = 1:numCls
    ind = find(labTe==i);
    temp = pred(ind);
    errCls(i) = length(find(temp~=i))/length(ind)*100;
end
% class-wise averaged error rate is used as evaluation metric for
% classification
err = mean(errCls);
disp(['classification error rate = ' num2str(err) '%'])

% Clustering on selection features (NMI and ACC)
addpath(genpath('nmi'))
disp('Clustering on selection features (NMI and ACC) ...')
numCls = length(unique(labTrSubset));
dataTeFeatSelect = dataTe(featKeepInd,:);
labTe(find(labTe==0)) = 10; % set digit 0 as the 10th class for evaluation
meanDataTeFeatSelect = zeros(numFeat,numCls);
for i = 1:numCls
    idx = find(labTe==i);
    x = dataTeFeatSelect(:,idx);
    meanDataTeFeatSelect(:,i) = mean(x,2)/norm(mean(x,2));
end
clusterACC = zeros(1,20);
clusterNMI = zeros(1,20);
for k = 1:20
    k
    rng(k)
    [idx,C] = kmeans(dataTeFeatSelect',numCls,'maxiter',400);
    for i = 1:numCls
        C(i,:) = C(i,:)/norm(C(i,:));
    end
    % ACC
    costMat = 1 - meanDataTeFeatSelect' * C';
    [assignment,cost] = munkres(costMat);
    labTePseudo = zeros(size(labTe));
    for i = 1:numCls
        ind = find(labTe==i);
        labTePseudo(ind) = assignment(i);
    end
    clusterACC(k) = mean([labTePseudo'==idx]);
    % NMI
    clusterNMI(k) = nmi(idx,labTe);
end
rmpath(genpath('nmi'))
ACC = mean(clusterACC);
disp(['clustering accuracy = ' num2str(ACC*100) '%'])
NMI = mean(clusterNMI);
disp(['normalized mutual information = ' num2str(NMI*100) '%'])